/*
 * Help Access Library 
 * A Library to access the contents of Windows Help files.
 *
 * Copyright (C) 1995-2000 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* SHGEXPOR.C -----------------------------------------
 * @doc
 * @module SHGEXPOR.C |
 *
 * Export-Support for SHG-Files.
 *
 * Converts DIBs and Htspot-Informations into SHG
 * (Segmented HyperGraphic) Files like the ones created
 * by Microsofts SHED or HC31.
 *
 * This function does not support creation of Metafile-based
 * SHG-Files.
 *
 * Help Access Library Project.
 *-----------------------------------------------------*/

#include <windows.h>
#include <windowsx.h>
#pragma hdrstop
#include <string.h>
#include <time.h>
#include "hlpacces.h"

#include "seg.h"
#include "shgimpor.h"           // SHG-Import API-Definition

#ifdef STATICLIB        /* Do not generate export-Definitions for static libraries */                        
# define DLLEXP
#else
# define DLLEXP _export
#endif


#ifdef __FLAT__
#  define HUGE
#else
#  define HUGE huge
#endif

#ifdef __FLAT__
# define hmemcpy  memcpy
# define hstrlen  lstrlen
//# define hstrcpyn lstrcpyn
#endif

typedef unsigned char HUGE *HPBYTE;
typedef          char HUGE *HPSTR;


DWORD static near pascal CompressRLE(HPBYTE input, HPBYTE output, DWORD NumBytes);


#pragma pack(1)
/* Hot Spot Header */
typedef struct tagHOTSPOTHEADER {
    BYTE  hhOne;           /* Always 0x01                                  */
    WORD  hhNumHS;         /* Number of Hot Spots                  */
    DWORD hhContextOffset; /* Offset to Cntxt Strings & IDs*/
}
PACKED HOTSPOTHEADER, FAR *LPHOTSPOTHEADER;

/* Hot Spot Record */
typedef struct tagHOTSPOTRECORD {
    WORD  hrType; /* Hot Spot Type. See below           */
    BYTE  hrZero; /* Always 0                                           */
    WORD  hrLeft; /* Bounding box of Hot Spot           */
    WORD  hrTop;
    WORD  hrRight;
    WORD  hrBottom;
    DWORD hrMacOffset;/* Offset to macro for Hot Spot   */
}
PACKED HOTSPOTRECORD, FAR *LPHOTSPOTRECORD;
#pragma pack()

#define HS_INVISJUMP            0x04E7
#define HS_INVISPOPUP           0x04E6
#define HS_INVISMACRO           0x04CC
#define HS_VISJUMP              0x00E3
#define HS_VISPOPUP             0x00E2
#define HS_VISMACRO             0x00C8


/* CreateHotspotData ---------------------------------------------
 *
 * Converts Hotspt-Data from SEGINFOHEADER into SHG-Format
 *
 *----------------------------------------------------------------*/
DWORD static near pascal CreateHotspotData(LPSTR pmem, HGLOBAL hSegInfo)
{ LPSEGINFOHEADER si   = (LPSEGINFOHEADER) GlobalLock(hSegInfo);
  LPSEGJUMP       sj;
  DWORD           size = sizeof(HOTSPOTHEADER) + si->SegJumpCount * sizeof(HOTSPOTRECORD);
  LPHOTSPOTHEADER hh   = (LPHOTSPOTHEADER) pmem;
  LPHOTSPOTRECORD hr;
  LPSTR           pMacro= pmem+size,
                  pSourceMacro;
  int             i;

  hh->hhOne   = 1;
  hh->hhNumHS = si->SegJumpCount;

  //------------ Transfer sizes and Macro-Strings --------
  for (i=0, sj= (LPSEGJUMP) (si+1), hr=(LPHOTSPOTRECORD) (hh+1);
       i<si->SegJumpCount;
       i++, sj++                  , hr++)
  {
    hr->hrType =  sj->flags & SI_POPUP ? HS_INVISPOPUP
                : sj->flags & SI_MACRO ? HS_INVISMACRO
                :                        HS_INVISJUMP;
    hr->hrLeft     = sj->left;
    hr->hrTop      = sj->top;
    hr->hrRight    = sj->right - sj->left;
    hr->hrBottom   = sj->bottom- sj->top;
    pSourceMacro = ((LPSTR) si) + sj->macro.dw;

    if (sj->flags & SI_MACRO)
    {
         hr->hrMacOffset= size-(sizeof(HOTSPOTHEADER)+ si->SegJumpCount * sizeof(HOTSPOTRECORD));
         lstrcpy(pMacro         , pSourceMacro);

         pMacro+= lstrlen(pSourceMacro)+1;
         size  += lstrlen(pSourceMacro)+1;
    }
    else
    {
         /* For other hotspots...use hrMacOffset for the Hash Code. */
         hr->hrMacOffset = HlpGenerateHashValue(pSourceMacro);
         pSourceMacro    = "";
    }
  }

  //------------ Transfer context-Strings ----------------
  hh->hhContextOffset = size-(sizeof(HOTSPOTHEADER)+ si->SegJumpCount * sizeof(HOTSPOTRECORD));

  for (i=0, sj= (LPSEGJUMP) (si+1), hr=(LPHOTSPOTRECORD) (hh+1);
       i<si->SegJumpCount;
       i++, sj++                  , hr++)
  {
    pSourceMacro = ((LPSTR) si) + sj->HotspotId.dw;

    lstrcpy(pMacro         , pSourceMacro);
            pMacro+= lstrlen(pSourceMacro)+1;
            size  += lstrlen(pSourceMacro)+1;

    pSourceMacro = ((LPSTR) si) + sj->macro.dw;

    lstrcpy(pMacro         , pSourceMacro);
            pMacro+= lstrlen(pSourceMacro)+1;
            size  += lstrlen(pSourceMacro)+1;
  }

  GlobalUnlock(hSegInfo);

  return size;
}



/* SHGExport -----------------------------------------------------
 * @func HGLOBAL | SHGExport |
 * Performs export of DIB as SHG-Graphic with Hotspots
 *
 * @xref <f SHGImport>
 *
 * @rvalue NULL | on Error
 * @rvalue !=NULL | Global Memory Handle containing the SHG file Data.
 *
 *----------------------------------------------------------------*/
HGLOBAL WINAPI DLLEXP SHGExport(
                #ifdef _Windows
                HWND             hwParent,        // @parm Parent-Window for error Messages
                #endif
                HGLOBAL          hDib,            // @parm DIB-Handle
                HGLOBAL          hSegInfo         // @parm Segmentation Informations or NULL if none available
                                                  //       See File SEG.H for more Informations on format
                )
{ DWORD              dwSize      = GlobalSize(hDib) + (hSegInfo ? GlobalSize(hSegInfo) : 0) + 0x8000;
  HGLOBAL            hmem        = GlobalAlloc(GHND, dwSize);
  LPSTR              pmem        = GlobalLock(hmem),
                     pdst        = pmem;
  LPBITMAPINFOHEADER lpbmi       = (LPBITMAPINFOHEADER) GlobalLock(hDib);
  UINT               BytesLine   = (lpbmi->biWidth * lpbmi->biBitCount + 31) / 32 * 4;
  DWORD              dwSizeDIB   =  lpbmi->biHeight* BytesLine,
                     dwSizeRLE, 
                     dwSizeHS;
  UINT               PaletteCount=  lpbmi->biClrUsed ? lpbmi->biClrUsed : (WORD) (1<<lpbmi->biBitCount);
  HGLOBAL            hHS         = GlobalAlloc(GHND, 0xf000);
  LPSTR              pHS         = (LPSTR) GlobalLock(hHS);

  if (pmem && pHS)
  {
    //------------------ First need to know exact Hotspot Size ------
    dwSizeHS = hSegInfo ? CreateHotspotData(pHS, hSegInfo) : 0;

    //------------------ Next need to know RLE-Compressed Size...----
    dwSizeRLE = CompressRLE((LPSTR)(lpbmi+1)+PaletteCount*sizeof(RGBQUAD), pmem, dwSizeDIB);

    //------------------ Create us a simple header ------------------
    memcpy(pdst, "lp\1\0"                       // mark as SHG Magic, 1 Object
                 "\x8\0\0\0"                    // Offset to Data
//                 "\6\0"                               //  Type 6:Bitmap Uncompressed,
                 "\6\1"                         //  Type 106:Bitmap RLEcompressed,
                 "\0\0\0\0", 14);               /* Don't tell anybody of the Resolution */
//                 "\xc0\0\xc0\0", 14);         /* Don't tell anybody of the Resolution */
    pdst += 14;

    /*------------------ Create the Bitmap-Header -------------------
    /* Read first three fields unmodified */


//    SHGBM->sbDPIX    = ReadDWordVal(&object); /* May be in "Dots Per Inch" or "Pixel per Meter" */
//    SHGBM->sbDPIY    = ReadDWordVal(&object);
                 * pdst ++=2;                                   // Number of Planes...
                 * pdst ++=lpbmi->biBitCount * 2;               // Number of Bits / Pixel

#   define PUTDWORD(val)  if ((DWORD) val > 0x7fff) \
                 *((DWORD FAR *)pdst)++=       val * 2+1; \
       else      *( (WORD FAR *)pdst)++=(WORD) val * 2;

    PUTDWORD(lpbmi->biWidth);
    PUTDWORD(lpbmi->biHeight);
    PUTDWORD(PaletteCount);
    PUTDWORD(lpbmi->biClrImportant);
    PUTDWORD(dwSizeRLE);
    PUTDWORD(dwSizeHS);
/*    *((DWORD FAR *)pdst)++=lpbmi->biWidth * 2+1;
    *((DWORD FAR *)pdst)++=lpbmi->biHeight* 2+1;
    *((DWORD FAR *)pdst)++=PaletteCount* 2+1;
    *((DWORD FAR *)pdst)++=lpbmi->biClrImportant * 2+1;
    *((DWORD FAR *)pdst)++=dwSizeDIB * 2+1;                     // Compressed Size
    *((DWORD FAR *)pdst)++=dwSizeHS  * 2+1;                     // Size of Hotspot-Data
*/
//    *((DWORD FAR *)pdst)++=(pdst-pmem+8)+PaletteCount*sizeof(RGBQUAD);// Picture Offset
//    *((DWORD FAR *)pdst)++=(pdst-pmem+4)+PaletteCount*sizeof(RGBQUAD)+dwSizeRLE;// Size of Image
    *((DWORD FAR *)pdst)++=(pdst-pmem  )+PaletteCount*sizeof(RGBQUAD);// Picture Offset
    *((DWORD FAR *)pdst)++=(pdst-pmem-4)+PaletteCount*sizeof(RGBQUAD)+dwSizeRLE;// Size of Image
    hmemcpy(pdst                             , lpbmi+1, PaletteCount*sizeof(RGBQUAD));
    dwSizeRLE = CompressRLE((LPSTR)(lpbmi+1)+PaletteCount*sizeof(RGBQUAD), pdst+PaletteCount*sizeof(RGBQUAD), dwSizeDIB);

    if (dwSizeHS)
       hmemcpy(((HPSTR) pdst)+PaletteCount*sizeof(RGBQUAD)+dwSizeRLE,   // Abbent Hotspot Data
             pHS, dwSizeHS);

    dwSize = (pdst-pmem) + PaletteCount*sizeof(RGBQUAD)+dwSizeRLE + dwSizeHS;

    GlobalUnlock(hmem);
    hmem = GlobalReAlloc(hmem, dwSize, GHND);
    
    GlobalUnlock(hHS); GlobalFree(hHS);
  }
  GlobalUnlock(hDib);

  return hmem;
}


/* Performs RLE compression for bitmaps */
DWORD static near pascal CompressRLE(HPBYTE input, HPBYTE output, DWORD NumBytes)
{
    UINT  equal, unequal;
    BYTE  val;
    DWORD outbytes=0;
    HPBYTE  stlen;

    while (NumBytes)
    {
       equal=1;
       val  =*input++;
       NumBytes--;
       while (val==*input && equal < 0x7f && NumBytes) { ++input; ++equal; NumBytes--; }

       if (equal>2)
       { *output++ = (BYTE) equal;
         *output++ = val;
         outbytes+=2;
       }
       else
       { unequal=equal;
         stlen=output++;
               outbytes++;
         while (equal--)
         {
              *output++=val;
               outbytes++;
         }
         while (((NumBytes > 1 && input[0]!=input[1]) ||
                 (NumBytes > 2 && input[1]!=input[2]))  
                && unequal<0x7f)
           { *output++=*input++;
              NumBytes--;
              outbytes++;
              unequal++;
           }
         *stlen = (BYTE) unequal | 0x80;
       }
    }   
    return outbytes;
}

