/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* wc_html.c -- part of WinHelpCGI

   This piece enumerates a topic of a HLP file and does the
   conversion into the HTML file format.
 
   (C)1996-2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/


#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hlpacces.h"           // Help Access Library Functions
#include "winhelpcgi.h"         // Header-File to connect the pieces of WinHelpCGI application

#include <libintl.h>

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)



/* The ENUMERATTION_PRIVATE_DATA structur contains fields that are needed by
   the callback-function during enumeration of a Topic. */
typedef struct
{
  BOOL     BrowseBackCalled;            /* TRUE after Title has been enumerated first */
  LPCSTR   HlpFileName;                 /* Filename of Help-File currently processed */
  DWORD    TopicOffset;                 /* TopicOffset of Topic currently processed */
  
  BOOL     Hidden;                      /* text is currently hidden (RTF \v)*/
  BOOL     InAnchor;                    /* An anchor-Tag <a ...> has is currently open */
  BOOL	  HtmlTable;                   /* A table-Tag <table> is currently open */
  BOOL	  HtmlRow;                     /* A Table-Row <tr><td> is currently open */
  BOOL	  RtfTable;                    /* RTF opened table with \intbl (will be closed by \row)
														 needed to collect several lines of a RTF-Tables to ONE
														 HTML-Table     
													 */
  BOOL	  InBrdrb;                     /* Display line below */        
  BOOL     FontSet;                     /* A font has been set */
  
  LOGFONT  lf[MAX_FONTS];               /* Font-Name from RTF \fonttbl */
  LOGFONT  lflast;                      /* Last font set by <font...> tag */
  LOGFONT  lfnow;                       /* Font that need to be set for next Character */
  RGBQUAD  cl[MAX_FONTS];               /* Color-Informations from RTF \colortbl */
  RGBQUAD  cllast;                      /* Last color set b< <font...> tag */
  RGBQUAD  clnow;                       /* Color that need to be set for next Character */

  HIFS     ifs;                         /* Access to internal file System */
  BOOL     InPara;                      /* Paragraph has been opened */
  BOOL     InNsr;                       /* Currently in a non-scrolling region */
  BOOL     NsrClosed;                   /* TRUE when </div> for closing NSR sent */
  int      rtf_sl;                      /* RTF-Parameter Line Spacing */
  int      rtf_sa;                      /* RTF-Parameter Spacing After */
  int      rtf_sb;                      /* RTF-Parameter Spacing Before */
  int      rtf_li;                      /* RTF-Parameter Left Indent */
  int      rtf_ri;                      /* RTF-Parameter Right Intent */
  int      rtf_fi;                      /* RTF-Parameter First Intent */
  int      BgColorNsr;                  /* Background-Color in Non-Scrolling Region */

  BOOL     InPopup;                     /* TRUE when displaying in a popup */
  BOOL     InPrint;                     /* Format for Printing */
} ENUMERATION_PRIVATE_DATA, FAR *LPENUMERATION_PRIVATE_DATA;


/* ------ Local Function prototypes ---------------------------------- */
void    htmlputs(LPCSTR src);
void    SimpleRtfToHtml(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszRtf);
static  void ResetFont(LPENUMERATION_PRIVATE_DATA dta);
static  void _OpenPara(LPENUMERATION_PRIVATE_DATA dta);

/* CreateAnchor
 *
 * Creates an anchor referencing to the Specified information.
 */
void CreateAnchor(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszTopicText)
{
  char  HlpFileName[MAX_PATH],
        WindowName[50],
        ContextId[50];
  LPSTR dst=ContextId;
  HIFS  NewIFS;
  DWORD ContextIdTopicOffset,
        TitleTopicOffset;

  WindowName[0]=0;
  HlpFileName[0]=0;

  _OpenPara(dta);

  /* Seperate Jump destination into Context-Id, Window-Name and Filename */
  strcpy(HlpFileName, dta->HlpFileName);

  while (*lpszTopicText)
  {
    switch (*lpszTopicText)
    {
       case '@': *dst=0; dst=HlpFileName; break;
       case '>': *dst=0; dst=WindowName; break;
       default : *dst++=*lpszTopicText;
    }
    lpszTopicText++;
  }
  *dst=0;

  /* Look if jump-Destination is currently available */
  NewIFS=IFSOpen(HlpFileName);
  if (NewIFS)
  {
     if (0xffffffffL != (ContextIdTopicOffset = HlpIsTopic(NewIFS, HlpGenerateHashValue(ContextId))) &&   // Context If can be found.
         HlpGetTopicTitle(NewIFS, ContextIdTopicOffset, NULL, 0, &TitleTopicOffset)&&                    // Need context-Id for start of Topic
         (TitleTopicOffset!=dta->TopicOffset ||                                                              // Do not refer inside of html-file else we'd have to use HTO_CONTEXT and this is slow...
          strcasecmp(HlpFileName, dta->HlpFileName))
        )
     {
       ResetFont(dta);

       /* Setup HTML anchor */
       if (WindowName[0]!=0 && strcmp(WindowName, "main"))
         {
           printf("<a href=\"" CGI_NAME "?FILENAME=%s&CONTEXTID=%s&FORMAT=popup\" target=\"%s\" onclick=\'window.open(\"" CGI_NAME "?FILENAME=%s&CONTEXTID=%s&FORMAT=popup\", \"%s\", \"width=400,height=400,screenX=0,screenY=0,dependent=yes,locationbar=no,menubar=no,status=no\")'>",HlpFileName,UrlEscape(ContextId), WindowName, HlpFileName, UrlEscape(ContextId), WindowName);
         }
       else if (dta->InPopup)
         {
           printf("<a href=\"" CGI_NAME "?FILENAME=%s&CONTEXT%s&FORMAT=popup=1\" target=\"main\" onclick=\'opener.location=\"" CGI_NAME "?FILENAME=%s&CONTEXTID=%s\"; return false;'>",HlpFileName, UrlEscape(ContextId), HlpFileName, UrlEscape(ContextId));
         }
       else
         {
           printf("<a href=\"" CGI_NAME "?FILENAME=%s&CONTEXTID=%s\">",HlpFileName, UrlEscape(ContextId));
           
         }
        dta->InAnchor=TRUE;
     }

     IFSClose(NewIFS);
  }
}


/**
 * Process a macro-string to create an anchor from it if 
 * It points to some Internet-Location, for example an EF("http://www.herdsoft.com/") macro
 *
 * @param dta             Enumeration object
 * @param lpszMacroString The Macro String.
 */
static
void
_ProcessMacroString(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszTopicText)
{
  const char *p;

  if (NULL!=(p=strstr(lpszTopicText,"http://")) ||
      NULL!=(p=strstr(lpszTopicText,"https://")) ||
      NULL!=(p=strstr(lpszTopicText,"mailto:")) )
    {
      char url[1000];
      char *dest;
      for (dest=url; *p && !strchr("\"'", *p) && (dest-url)<sizeof(url)-1; )
        *dest++=*p++;
      *dest=0;

      printf("<a href=\"%s\">", url);
      dta->InAnchor=TRUE;
    }
}





/* ResetFont
 *
 * Reset font settings.
 */
static
void
ResetFont(LPENUMERATION_PRIVATE_DATA dta)
{
  if (dta->FontSet)
	 {
		if (dta->lflast.lfWeight>FW_NORMAL)                              printf("</b>");
		if (dta->lflast.lfItalic)                                        printf("</i>");
		if (dta->lflast.lfUnderline)                                     printf("</u>");
		
		puts("</font>");
		
		dta->FontSet = FALSE;
	 }
} // ResetFont


/* SetFont
 *
 * Set the font according to the selected parameters 
 */
void
SetFont(LPENUMERATION_PRIVATE_DATA dta)
{
  ResetFont(dta);

  printf("<font size=\"%d\" face=\"%s\"",
			dta->lfnow.lfHeight/5,
			dta->lfnow.lfFaceName
			);

  if (!dta->InAnchor && (dta->clnow.rgbRed>1||dta->clnow.rgbGreen>1||dta->clnow.rgbBlue>1))
    {
      printf(" color=\"#%02x%02x%02x\"",
             dta->clnow.rgbRed, dta->clnow.rgbGreen, dta->clnow.rgbBlue);
    }

  puts(">");

  if (dta->lfnow.lfWeight>FW_NORMAL)                               printf("<b>");
  if (dta->lfnow.lfItalic)                                         printf("<i>");
  if (dta->lfnow.lfUnderline)                                      printf("<u>");

  dta->cllast = dta->clnow;
  dta->lflast = dta->lfnow;

  dta->FontSet = TRUE;
  //  memcpy(&dta->cllast, &dta->clnow, sizeof(RGBQUAD));
  //memcpy(&dta->lflast, &dta->lfnow, sizeof(LOGFONT));
} // SetFont


/**
 * Open a Paragraph (if not already done) and set Paragraph style parameters
 */
static
void
_OpenPara(LPENUMERATION_PRIVATE_DATA dta)
{
  if (!dta->InPara)
    {
      if (!dta->InNsr && !dta->NsrClosed)
        {
          dta->NsrClosed = TRUE;
          fputs("</div>", stdout);
        }

      fputs("<p style=\"", stdout);
      if (dta->rtf_sa != 0)
        {
          printf("margin-bottom:%.1fpt;", dta->rtf_sa / 20.0);
        }
      if (dta->rtf_sl != 0)
        {
          printf("line-height:%.1fpt;", dta->rtf_sl / 20.0);
        }
      if (dta->rtf_sb != 0)
        {
          printf("margin-top:%.1fpt;", dta->rtf_sb / 20.0);
        }
      if (dta->rtf_li != 0)
        {
          printf("margin-left:%.1fpt;", dta->rtf_li / 20.0);
        }
      if (dta->rtf_ri != 0)
        {
          printf("margin-right:%.1fpt;", dta->rtf_ri / 20.0);
        }
      if (dta->rtf_fi != 0)
        {
          printf("text-indent:%.1fpt;", dta->rtf_fi / 20.0);
        }
      if (dta->InNsr && dta->BgColorNsr!=0xffffff)
        {
          printf("background-color: #%6x", dta->BgColorNsr);
        }

      fputs("\">", stdout);
      
      dta->InPara=TRUE;
    }
}


/**
 * Close a paragraph.
 */
static
void
_ClosePara(LPENUMERATION_PRIVATE_DATA dta)
{
  if (dta->InPara)
    {
      dta->InPara = FALSE;
      
      ResetFont(dta);

      puts(dta->InBrdrb ? "<hr>\n" : "</p>\n");
    }
}



/* CreateImageRef
 *
 * Creates an reference to an image
 */
void CreateImageRef(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszTopicText)
{
  LPCSTR p=strchr(lpszTopicText, ' '),
         szAlign="";

  switch (*(p-1))
  {
    case 'l' : szAlign=" align=left"; break;
    case 'r' : szAlign=" align=right"; break;
  }

  p++;
  *(strchr(p,'}'))=0;

  _OpenPara(dta);

  printf("<img src=\"" CGI_NAME "?FILENAME=%s&IMAGE=%s\"%s>",dta->HlpFileName,p,szAlign);
} // CreateImageRef


/* TopicEnumerationLinkRelProcedure
 *
 * This is the callback-Function this application defined for the call to
 * HlpTopicEnumerate to Enumerate the TOPIC-Files contents
 *
 * This version is just responsible to create the <link rel=...> tags in the header
 */
BOOL 
FAR PASCAL 
TopicEnumerationLinkRelProcedure(
                                 LPCSTR    lpszTopicText,   // Text to be transfered to the Application program
                                 DWORD     dummy,           // resevred
                                 TEXTTYPE  eTextType,       // type of Text-String transfered to the Application
                                 LPARAM    lParam)          // Application-Defined user Data
{
  BOOL result = TRUE;
  LPENUMERATION_PRIVATE_DATA dta = (LPENUMERATION_PRIVATE_DATA) lParam;
  
  switch (eTextType)
	 {
	 case TOPIC_BROWSE_BACK:
	 case TOPIC_BROWSE_FORWARD:
		{
		  DWORD dwTopicOffset;
		  char  szTopicTitle[512];
		  sscanf(lpszTopicText, "%8x", &dwTopicOffset);
		  
		  if (dwTopicOffset!=0xffffffffL)
			 {
				*szTopicTitle=0;
				HlpGetTopicTitle(dta->ifs, dwTopicOffset, szTopicTitle, sizeof(szTopicTitle)-1, NULL);
				
				printf("<link rel=\"%s\" href=\"" CGI_NAME "?FILENAME=%s&TOPICOFFSET=0x%s\" title=\"",
						 eTextType==TOPIC_BROWSE_FORWARD ? "next" : "prev",
						 dta->HlpFileName,
						 lpszTopicText);
				htmlputs(szTopicTitle);
            puts("\">");
			 }

        if (eTextType==TOPIC_BROWSE_BACK)
          {
            if (!dta->BrowseBackCalled)
              {
                dta->BrowseBackCalled=TRUE;
          }
            else
              {
                result = FALSE;
              }
          }
        break;

		}
		break;
	 case TOPIC_TEXT:                   // Normal Text...
      result = FALSE;
      break;

    default: ;
    }

  return result;
}


/* TopicEnumerationProcedure
 *
 * This is the callback-Function this application defined for the call to
 * HlpTopicEnumerate to Enumerate the TOPIC-Files contents
 */
BOOL 
FAR PASCAL 
TopicEnumerationProcedure(
								  LPCSTR    lpszTopicText,   // Text to be transfered to the Application program
								  DWORD     dummy,           // resevred
								  TEXTTYPE  eTextType,       // type of Text-String transfered to the Application
								  LPARAM    lParam)          // Application-Defined user Data
{
  LPENUMERATION_PRIVATE_DATA dta = (LPENUMERATION_PRIVATE_DATA) lParam;
  
  switch (eTextType)
	 {
	 case TOPIC_RTF_INIT:                     // RTF Control strings
	 case TOPIC_RTF_TEXT:                     // RTF Control strings
	 case TOPIC_RTF_EXIT:                     // RTF Control strings
		SimpleRtfToHtml(dta, lpszTopicText);
		break;
		
	 case TOPIC_TEXT:                   // Normal Text...
		
		if (dta->HtmlTable && !dta->RtfTable)
		  {  
			 /* Theoretically a Table always spans only one line in RTF, so we have to collect them */
			 dta->HtmlTable=FALSE;
			 
			 /* Always close font when changing table items */
			 _ClosePara(dta);

			 if (dta->HtmlRow)
				{
				  puts("</td></tr>");
				}
			 dta->HtmlRow=FALSE;
			 puts("</table>");
		  }
		
		if (dta->HtmlTable && !dta->HtmlRow)
		  {
			 /* Always close font when changing table items */
			 _ClosePara(dta);

			 puts("<tr><td valign=top>");
			 dta->HtmlRow=TRUE;
		  }
		

      if (*lpszTopicText)
        {
          _OpenPara(dta);
        }
		if (memcmp(&dta->clnow,&dta->cllast,sizeof(RGBQUAD))||
			 memcmp(&dta->lfnow,&dta->lflast,sizeof(LOGFONT))||
			 !dta->FontSet)
		  {
			 if (*lpszTopicText)
				{
				  SetFont(dta);
				}
		  }
		
		htmlputs(lpszTopicText);
		break;
		
	 case TOPIC_BROWSE_BACK:
	 case TOPIC_BROWSE_FORWARD:
		{
		  DWORD dwTopicOffset;
		  char  szTopicTitle[512];
        const char *makeBorder = " style=\"width:80px; text-align:center; border-width:medium; border-color:#808080; background-color: #c0c0c0; border-style:outset\"";

        if (eTextType==TOPIC_BROWSE_BACK)
          {
            if (!dta->BrowseBackCalled)
              {
                dta->BrowseBackCalled=TRUE;
              }
            else
              {
                return FALSE;
              }
          }

		  if (!dta->InPopup)
          {
            if (eTextType==TOPIC_BROWSE_BACK)
              {
                printf("<p><a href=\"" CGI_NAME "\"%s title=\"%s\">%s</a> ",
                       makeBorder,
                       _("List of available help files"),
                       _("Directory"));
                printf("<a href=\"" CGI_NAME "?FILENAME=%s\"%s title=\"%s\">%s</a> ",
                       dta->HlpFileName,
                       makeBorder,
                       _("Display the contents listing of this Help File"),
                       _("Contents"));
                printf("<a href=\"" CGI_NAME "?FILENAME=%s&KEYWORD=\"%s title=\"%s\">%s</a> ",
                       dta->HlpFileName,
                       makeBorder,
                       _("Search the current help file for keywords"),
                       _("Search")
                       );
                printf("<a href=\"" CGI_NAME "?FILENAME=%s&TOPICOFFSET=0x%08x&FORMAT=print\"%s title=\"%s\" target=\"print\">%s</a> ",
                       dta->HlpFileName,
                       dta->TopicOffset,
                       makeBorder,
                       _("Print the current topic"),
                       _("Print")
                       );
              }
            
            sscanf(lpszTopicText, "%8x", &dwTopicOffset);
            
            if (dwTopicOffset!=0xffffffffL)
              {
                *szTopicTitle=0;
                HlpGetTopicTitle(dta->ifs, dwTopicOffset, szTopicTitle, sizeof(szTopicTitle)-1, NULL);
                
                printf("<a href=\"" CGI_NAME "?FILENAME=%s&TOPICOFFSET=0x%s\"%s title=\"",
                       dta->HlpFileName,
                       lpszTopicText,
                       makeBorder);
                htmlputs(szTopicTitle);
                printf("\">%s</a> ",
                       eTextType==TOPIC_BROWSE_FORWARD ? "&gt;&gt;" : "&lt;&lt;");
              }
            
            if (eTextType==TOPIC_BROWSE_FORWARD)
              {
                printf("</p>"
                       "<hr>");

                if (!dta->InPopup)
                  {
                    InsertInclude(dta->HlpFileName, "text", NULL);
                  }
              }
          }
		}
		break;
		
	 case TOPIC_HOTSPOT_START:          // reference to another topic or Anchor
		if (*lpszTopicText=='%' || // Invisible Hotspots...
			 *lpszTopicText=='*')
		  lpszTopicText++;

      CreateAnchor(dta, lpszTopicText);
		break;

    case TOPIC_MACRO_HOTSPOT:
		if (*lpszTopicText=='%' || // Invisible Hotspots...
			 *lpszTopicText=='*')
		  lpszTopicText++;

      _ProcessMacroString(dta, lpszTopicText);
        
      break;
		
	 case TOPIC_HOTSPOT_END:            // reference to another topic or Anchor END
		if (dta->InAnchor)
		  puts("</a>");
		dta->InAnchor=FALSE;
		break;
		
	 case TOPIC_BMX:                    // Reference to a Bitmap (or Metafile :-( )
		CreateImageRef(dta, lpszTopicText);

    default: ;
	 }
  
  return TRUE;                          // Continue enumeration
}




/* EnumerateTopic
 *
 * Enumerates the Contents of one Topic of a Help File.
 *
 * @param HlpFileName      Filename of the .hlp or .mvb file.
 * @param HlpTopicOffset   Topic-Offset stored as a string
 * @param HlpContextId     Context-ID of the topic that shall be displayed
 * @param HlpFormat        Formatting-Parameter, "popup", "print" or NULL
 *
 */
void EnumerateTopic(LPCSTR HlpFileName, LPCSTR HlpTopicOffset, LPCSTR HlpContextId, LPCSTR HlpFormat)
{ 
  HIFS                     ifs= IFSOpen(HlpFileName);
  HTOPIC                   topic;
  DWORD                    TopicOffset=0;
  LPENUMERATION_PRIVATE_DATA lpdta=(LPENUMERATION_PRIVATE_DATA) calloc(sizeof(ENUMERATION_PRIVATE_DATA), 1);
  char                     szTopicTitle[512];
  int                      BgColor=0xffffff;

  puts(   "Content-type: text/html\n\n"
          "<html><head>");

  if (ifs && lpdta)
	 {
      if (HlpTopicOffset!=NULL)
        {
          sscanf(HlpTopicOffset+2, "%x", &TopicOffset);
        }
      else if (HlpContextId!=NULL)
        {
          TopicOffset = HlpIsTopic(ifs, HlpGenerateHashValue(HlpContextId));
        }
		
      /* ------------- Inclue Topic Title as HTML-Title ------------------*/
      printf("<title>");

      {
        LPHELPFILETITLE lpHelpTitle;
        lpHelpTitle = HlpGetHelpTitle(ifs);
        if (lpHelpTitle )
          {
            if (lpHelpTitle->HelpTitle)
              {
                htmlputs(lpHelpTitle->HelpTitle);
                printf(" - ");
              }
            BgColor = lpHelpTitle->BgColor;
            lpdta->BgColorNsr = lpHelpTitle->BgColorNsr;
            HlpFreeHelpTitle(lpHelpTitle);
          }
      }


      HlpGetTopicTitle(ifs, TopicOffset, szTopicTitle, sizeof(szTopicTitle), &TopicOffset);
      htmlputs(szTopicTitle);
      puts("</title>");

      printf("<link rel=\"up\" href=\"" CGI_NAME "?FILENAME=%s\" title=\"Contents\">\n", HlpFileName);
      printf("<link rel=\"start\" href=\"" CGI_NAME "\" title=\"Start\">\n");
      printf("<link rel=\"bookmark\" href=\"http://www.herdsoft.com/\" title=\"Herd Software Development\">\n");
      printf("<link rel=\"bookmark\" href=\"http://www.herdsoft.com/catalog/ehlp2rtf.html\" title=\"Help to RTF Converter\">\n");
      printf("<link rel=\"bookmark\" href=\"http://www.herdsoft.com/linux/produkte/winhelpcgi.html\" title=\"WinHelpCGI project page\">\n");
      printf("<link rel=\"search\" href=\"" CGI_NAME "?FILENAME=%s&KEYWORD=\" title=\"Glossary\">\n", HlpFileName);

      /* ------------- Enumerate the Topic Contents ----------------------*/
      topic = HlpTopicOpen(ifs, HTO_FONTS /*| HTO_CONTEXT | HTO_KEYWORDS*/);          // Open TOPIC-File and load Phrases etc.

      if (topic)
		  {
			 lpdta->BrowseBackCalled = FALSE;
			 lpdta->HlpFileName     = HlpFileName;
			 lpdta->TopicOffset     = TopicOffset;
			 lpdta->ifs             = ifs;
          lpdta->InPopup         = HlpFormat != NULL;
          lpdta->InPrint         = HlpFormat != NULL && !strcasecmp(HlpFormat, "print");

          if (!lpdta->InPopup)
            {
              InsertInclude(lpdta->HlpFileName, "head", NULL);
            }
  
          //------------ Create <link....> tags in header
          if (TopicOffset < 0 ||
              HlpTopicSeek(topic, TopicOffset) == -1)
            {
              printf("<p>TopicOffset is beyond end of File.</p>");
              TopicOffset = 0;
              HlpTopicSeek(topic, TopicOffset);
            }

			 HlpTopicEnumerate(topic, TopicEnumerationLinkRelProcedure, (LPARAM)lpdta);
			 lpdta->BrowseBackCalled = FALSE;

          printf("<style type=\"text/css\">\n"
                 "<!--\n"
                 " p { margin-bottom:0pt; margin-top:0pt; }\n"
                 "//-->\n"
                 "</style>\n\n");

          puts("</head>");

          if (!lpdta->InPopup)
            {
              char body[500];
              snprintf(body, sizeof(body), "<body bgcolor=\"#%6x\">", BgColor);
              InsertInclude(lpdta->HlpFileName, "body", body);
            }

			 printf("<font size=\"3\">");        // open ANY font-Tag
			 
			 //------------ Enumerate the Topic ------------------------------
			 HlpTopicSeek(     topic, TopicOffset);
          printf("<div style=\"background-color: #%6x;\"><p>", lpdta->BgColorNsr);
			 HlpTopicEnumerate(topic, TopicEnumerationProcedure, (LPARAM)lpdta);
			 HlpTopicClose(topic);
			 
			 puts("</font>");                // close ANY font-Tag
		  }
      else
        printf(ERROR_FONT "</head><body>HlpTopicOpen: Topic-File could not be opened. Maybe it is not a Help-File?\n" NORMAL_FONT);
		
      if (!lpdta->InPopup)
        {
          InsertInclude(lpdta->HlpFileName, "endtext", NULL);
        }
      puts("</body></html>");

      if (lpdta->InPrint)
        {
          puts("<script language=\"JavaScript\">\n"
               "<!--\n"
               "  window.print();\n"
               "-->\n"
               "</script>");
        }
          
		IFSClose(ifs);

	 } else printf(ERROR_FONT "Sorry, can't open File %s" NORMAL_FONT, HlpFileName);

  if (lpdta)
    {
      free(lpdta);
    }
} // EnumerateTopic



/* htmlputs ----------------------------------------------------*
 * Display a string that may contain special Characters as HTML
 *--------------------------------------------------------------*/
void htmlputs(LPCSTR src)
{ 
  char * block = malloc(strlen(src)*6+1000);
  char * dst;
  
  if (block)
	 {
		for (dst=block;
			  *src;
			  ++src)
		  switch (((unsigned char) *src) & 0xff)
			 {
			 case 0xe4: strcpy(dst, "&auml;"); dst+=6; break;
			 case 0xf6: strcpy(dst, "&ouml;"); dst+=6; break;
			 case 0xfc: strcpy(dst, "&uuml;"); dst+=6; break;
			 case 0xc4: strcpy(dst, "&Auml;"); dst+=6; break;
			 case 0xd6: strcpy(dst, "&Ouml;"); dst+=6; break;
			 case 0xdc: strcpy(dst, "&Uuml;"); dst+=6; break;
			 case 0xdf: strcpy(dst, "&szlig;"); dst+=7; break;
			 case '<' : strcpy(dst, "&lt;"); dst+=4; break;
			 case '>' : strcpy(dst, "&gt;"); dst+=4; break;
			 case '&' : strcpy(dst, "&amp;"); dst+=5; break;
			 case '\"': strcpy(dst, "&quot;"); dst+=6; break;
          case  183: strcpy(dst, "&bull;"); dst+=6; break; // a bullet in Symbol font
          case  222: strcpy(dst, "&rarr;"); dst+=6; break; // a double right arrow in Symbol font
			 default:   if (*src & 0x80)
				{
				  sprintf(dst, "&#%d;", ((unsigned char) *src) & 0xff);
				  dst+=strlen(dst);
				}
			 else
				if (*src >= ' ')
				  *dst++=*src;
			 }
		*dst=0;
		
		fwrite(block, 1, strlen(block), stdout);
		
		free(block);
	 }
} // htmlputs




LPCSTR 
SimpleReadRtfCommand(LPCSTR lpszRtf, LPSTR dst, LPSTR par);


LPCSTR 
SimpleReadFontTable(LPCSTR lpszRtf, LPLOGFONT lf)
{
  int   count=0;
  char *dst;

  while (*lpszRtf && *lpszRtf!='}' && count<MAX_FONTS)
	 {
		while (*lpszRtf && *lpszRtf!='{' && *lpszRtf!='}') ++lpszRtf;
		
		if (*lpszRtf=='{')
		  {
			 while (*lpszRtf && *lpszRtf!=' ') ++lpszRtf;
			 
			 if (*lpszRtf==' ')
				{  lpszRtf++;
				for (dst=lf[count++].lfFaceName;
					  *lpszRtf && *lpszRtf!=';';
					  )
				  *dst++=*lpszRtf++;
				*dst=0;
				
				}
			 
			 if (*lpszRtf==';') lpszRtf++;
			 if (*lpszRtf=='}') lpszRtf++;
		  }
	 }
  
  return lpszRtf;
} // SimpleReadFontTable


LPCSTR 
SimpleReadColorTable(LPCSTR lpszRtf, LPRGBQUAD cl)
{
  int   count=0;
  char  szRtfCommand[100],
        szRtfPar[100];

  while (*lpszRtf && *lpszRtf!='}' && count<MAX_FONTS-1)
  {
    switch (*lpszRtf)
    {
      case ';':  { count++; ++lpszRtf; } break;
      case '\\': lpszRtf=SimpleReadRtfCommand(lpszRtf+1, szRtfCommand, szRtfPar);
                      if (!stricmp(szRtfCommand,"red"  )) cl[count].rgbRed  =(BYTE)atoi(szRtfPar);
                 else if (!stricmp(szRtfCommand,"green")) cl[count].rgbGreen=(BYTE)atoi(szRtfPar);
                 else if (!stricmp(szRtfCommand,"blue" )) cl[count].rgbBlue =(BYTE)atoi(szRtfPar);
                 break;
      default:   lpszRtf++;
    }
  }

  return lpszRtf;
} // SimpleReadColorTable


LPCSTR 
SimpleReadRtfCommand(LPCSTR lpszRtf, LPSTR szRtfCommand, LPSTR szRtfPar)
{   
  char * dst;

  for  ( dst=szRtfCommand;
			(*lpszRtf>='a' && *lpszRtf<='z')||
           (*lpszRtf>='A' && *lpszRtf<='Z')||
			  *lpszRtf=='~';
         )
	 *dst++=*lpszRtf++;
  *dst=0;
  
  for  ( dst=szRtfPar;
			(*lpszRtf>='0' && *lpszRtf<='9')||
			  *lpszRtf=='-';
         )
	 *dst++=*lpszRtf++;
  *dst=0;
  
  if (*lpszRtf && strchr("\r\n\t ", *lpszRtf))
	 ++lpszRtf;
  
  return lpszRtf;
} // SimpleReadRtfCommand


/* SimpleRTfToHtml
 *
 * This routine parses the RTF-Codes transfered by HlpEnumTopic callback
 * to find out about fonts, tables etc. The Coding only supports the
 * subset of RTF used by Help Access Library, it would fail on a
 * RTF-File created by a Word-Processing application, mainly due
 * to the fact that we haven't supported restaurtaion of font-
 * attrubutes on { } parantheses.
 */
void 
SimpleRtfToHtml(LPENUMERATION_PRIVATE_DATA dta, LPCSTR lpszRtf)
{
  char  szRtfCommand[100];
  char  szRtfPar[100];
  
  while (*lpszRtf)
	 {
		if (*lpszRtf=='\\')
		  {
			 lpszRtf=SimpleReadRtfCommand(lpszRtf+1, szRtfCommand, szRtfPar);
		  
			 if (*lpszRtf=='}') dta->Hidden=FALSE;
			 
			 if (!stricmp(szRtfCommand,"b"))    dta->lfnow.lfWeight=szRtfPar[0]=='0'?FW_NORMAL:FW_BOLD;
			 else if (!stricmp(szRtfCommand,"v"))    dta->Hidden=szRtfPar[0]!='0';
			 else if (!stricmp(szRtfCommand,"ul"))   dta->lfnow.lfUnderline=szRtfPar[0]!='0';
			 else if (!stricmp(szRtfCommand,"i"))    dta->lfnow.lfItalic=szRtfPar[0]!='0';
			 else if (!stricmp(szRtfCommand,"line")) printf("<br>\n");
			 else if (!stricmp(szRtfCommand,"tab"))  printf("&nbsp;&nbsp;");
			 else if (!stricmp(szRtfCommand,"fonttbl"))  lpszRtf=SimpleReadFontTable(lpszRtf, dta->lf);
			 else if (!stricmp(szRtfCommand,"colortbl")) lpszRtf=SimpleReadColorTable(lpszRtf, dta->cl);
			 else if (!stricmp(szRtfCommand,"fs"))       dta->lfnow.lfHeight=atoi(szRtfPar);
			 else if (!stricmp(szRtfCommand,"cf"))       dta->clnow=dta->cl[atoi(szRtfPar)];
			 else if (!stricmp(szRtfCommand,"f" ))       strcpy(dta->lfnow.lfFaceName, dta->lf[atoi(szRtfPar)].lfFaceName);
			 else if (!stricmp(szRtfCommand,"brdrb"))    dta->InBrdrb=TRUE;
			 else if (!stricmp(szRtfCommand,"brdrt"))    puts("<hr>");
			 else if (!stricmp(szRtfCommand,"~"))        puts("&nbsp");
			 
			 else if (!stricmp(szRtfCommand,"intbl" ))   
				{ 
				  if (!dta->HtmlTable)
					 { 
						dta->HtmlTable=TRUE;

						/* Always close font when changing table items */
						ResetFont(dta);

						puts("<table border=\"0\">");
					 }
				  dta->RtfTable=TRUE;
				}
			 else	if (!stricmp(szRtfCommand,"row" ))     
            { 
              /* Always close font when changing table items */
              ResetFont(dta);
              
              dta->RtfTable=FALSE;
              puts("</td></tr>");
              dta->HtmlRow=FALSE;
            }
          else if (!stricmp(szRtfCommand,"cell" ))    
            {
              /* Always close font when changing table items */
              _ClosePara(dta);
              
              puts("</td><td valign=top>");
            }
          else if (!stricmp(szRtfCommand,"sa"))
            {
              dta->rtf_sa = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"sl"))
            {
              dta->rtf_sl = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"sb"))
            {
              dta->rtf_sb = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"li"))
            {
              dta->rtf_li = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"ri"))
            {
              dta->rtf_ri = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"fi"))
            {
              dta->rtf_fi = atoi(szRtfPar);
            }
          else if (!stricmp(szRtfCommand,"par"))
            {
              /* End of Paragraph */
              _ClosePara(dta);
            }
          else if (!stricmp(szRtfCommand, "pard"))
            {
              /* Reset Paragraph settings back to the default */
              dta->InBrdrb=FALSE;

              dta->rtf_li =
              dta->rtf_ri =
              dta->rtf_fi =
              dta->rtf_sl =
              dta->rtf_sa =
              dta->rtf_sb = 0;
              
              dta->InNsr = FALSE;

              _ClosePara(dta);
            }
          else if (!stricmp(szRtfCommand, "keepn"))
            {
              dta->InNsr = TRUE;
            }
		  }
      else
		  {
			 if (!strchr("\r\n\t{}",*lpszRtf) && !dta->Hidden)
				{
              _OpenPara(dta);

              szRtfCommand[0] = *lpszRtf;
				  szRtfCommand[1] = 0;
				  htmlputs(szRtfCommand);
				  lpszRtf++;
				}
			 else
				lpszRtf++;
		  }
	 }
} // SimpleRtfToHtml


/*
 * Local variables:
 *  tab-width: 3
 * End:
 */
