/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2000 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* wc_kwsrch.c -- part of WinHelpCGI

   Searches in all help files for a specified keyword and
   outputs an overview in html format.
 
   (C)1996-2003 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/


#include <windows.h>
#include <dirent.h>
/*#include <fcntl.h>
#if defined(__MSDOS__) || defined(__WIN32__)
#  include <io.h>
#else
#  include <dirent.h>
#endif
#include <time.h>*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hlpacces.h"           // Help Access Library Functions
#include "winhelpcgi.h"         // Header-File to connect the pieces of WinHelpCGI application

#include <libintl.h>

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)


typedef struct
{
  LPCSTR  lpcszTopicTitle;      /* Title of the Topic containing the KeyWord */
  LPCSTR  lpcszHelpFileName;    /* Anchor-String used to encode the link */
  DWORD   dwTopicOffset;        /* Topic Offset in that help file. */
} TOPICREF, *LPTOPICREF;

typedef struct 
{
  LPCSTR  lpcszKeyWord;         /* KeyWord encountered. */
  int     iHitCount;            /* Number of Hits for that KeyWord */
  LPTOPICREF lpTopicRefs;        /* Pointer to an array of TopicRef-Structure accessing containing this KeyWord */
} KWREF, *LPKWREF;

static int     iKwRefCount=0;       /* Number of Entries in lpKwRef */
static LPKWREF lpKwRefArray=NULL;

/* kwRefCompareFunction
 *
 * Function used to scan for KWREFs
 */
int kwRefCompareFunction(const void *a, const void *b)
{
  LPKWREF pa=(LPKWREF)a,
          pb=(LPKWREF)b;

  return strcasecmp(pa->lpcszKeyWord, pb->lpcszKeyWord);
} // kwRefCompareFunction


/* FindKWRef
 * 
 * Find KWREF entry matching th e current KeyWord or create one if this
 * is the first reference
 */
LPKWREF FindKWRef(LPCSTR lpcszKeyWord)
{
  KWREF     kwRefCompareValue;
  LPKWREF   lpkwRef;

  kwRefCompareValue.lpcszKeyWord=lpcszKeyWord;

  lpkwRef = iKwRefCount
          ? bsearch(&kwRefCompareValue, lpKwRefArray, iKwRefCount, sizeof(KWREF), kwRefCompareFunction)
          : NULL;

  if (!lpkwRef)
    {

      lpKwRefArray = realloc(lpKwRefArray, ++iKwRefCount * sizeof(KWREF));
      lpkwRef = lpKwRefArray+(iKwRefCount-1);
      lpkwRef->lpcszKeyWord = strdup(lpcszKeyWord);
      lpkwRef->iHitCount    = 0;
      lpkwRef->lpTopicRefs  = NULL;

      qsort(lpKwRefArray, iKwRefCount, sizeof(KWREF), kwRefCompareFunction);
      lpkwRef =  bsearch(&kwRefCompareValue, lpKwRefArray, iKwRefCount, sizeof(KWREF), kwRefCompareFunction);
    }

  return lpkwRef;
} // FindLWRef


/* AddTopicToKwRef
 *
 * Add a topic title and Reference to the list of topics associated to a KeyWord
 */
void static AddTopicToKwRef(
     LPKWREF lpKwRef, 
     LPCSTR  lpcszTopicTitle, 
     LPCSTR  lpcszHelpFileName, 
     DWORD   dwTopicOffset)
{
  LPTOPICREF lpTopicRef;

  lpKwRef->lpTopicRefs = realloc(lpKwRef->lpTopicRefs, ++lpKwRef->iHitCount * sizeof(TOPICREF));
  lpTopicRef = lpKwRef->lpTopicRefs + (lpKwRef->iHitCount-1);
  lpTopicRef->lpcszTopicTitle  =strdup(lpcszTopicTitle); 
  lpTopicRef->lpcszHelpFileName=lpcszHelpFileName;
  lpTopicRef->dwTopicOffset    =dwTopicOffset;

} // AddTopicToKwRef




typedef struct 
{
  LPCSTR    lpcszKeyWord;          /* KeyWord to search for */
  LPCSTR    lpcszHelpFileName;     /* Name of the Help File being analyzed */
  HIFSFILE  hKWData;               /* Handle of |KWDATA internal file. */
  HIFS      ifs;
} KWENUMPRIVATEDATA, *LPKWENUMPRIVATEDATA;


        BOOL FAR PASCAL KWSearchEnumProc(LPCSTR lpszKeyWord, DWORD dwDataOffset, UINT uDataCount, LPARAM lParam)
        {
          LPKWENUMPRIVATEDATA tKWList=(LPKWENUMPRIVATEDATA)lParam;
          int       i;
          DWORD     dwTopicOffset, dwTopicOffsetStart;
          BYTE      abTopicOffset[4];
          char      szTopicTitle[1024];
          BOOL      bGoodKeyWord;
          LPKWREF   lpKWRef;

          bGoodKeyWord = !strncasecmp(lpszKeyWord, tKWList->lpcszKeyWord, strlen(tKWList->lpcszKeyWord));

//          printf("Offset:0x%08lx Count:0x%04x KW:%s\n", dwDataOffset, uDataCount, lpszKeyWord);

          if (bGoodKeyWord &&
             NULL!=(lpKWRef=FindKWRef(lpszKeyWord)))
            {
              IFSSeekFile(tKWList->hKWData, dwDataOffset, SEEK_SET);

              for (i=0; i<uDataCount; ++i)
                {
                  IFSReadFile(tKWList->hKWData, &abTopicOffset, sizeof(abTopicOffset));

                  dwTopicOffset = abTopicOffset[0]+(abTopicOffset[1]<<8)+(abTopicOffset[2]<<16)+(abTopicOffset[3]<<24);

                  if (HlpGetTopicTitle(tKWList->ifs, dwTopicOffset, szTopicTitle, sizeof(szTopicTitle), &dwTopicOffsetStart))
                    {
                      AddTopicToKwRef(lpKWRef, szTopicTitle, tKWList->lpcszHelpFileName, dwTopicOffset);
                    }
                  //            else printf("  TopicOffset: 0x%08x: ERROR\n\n", dwTopicOffset);
                }
            }


          return bGoodKeyWord;
        }

   


/* CollectInFile
 *
 * Collect KeyWords from a specific Help File 
 */
void CollectInFile(HIFS ifs, LPCSTR lpcszHelpFileName, LPCSTR lpcszKeyWord)
{
  KWENUMPRIVATEDATA kwEnumPrivateData;

  kwEnumPrivateData.lpcszHelpFileName = strdup(lpcszHelpFileName);
  kwEnumPrivateData.lpcszKeyWord      = lpcszKeyWord;
  kwEnumPrivateData.hKWData           = IFSOpenFile(ifs, "|KWDATA");
  kwEnumPrivateData.ifs               = ifs;

  if (kwEnumPrivateData.hKWData)
    {
      HlpEnumKeyWords(ifs, KWSearchEnumProc, lpcszKeyWord, "|KWBTREE", (LPARAM)(LPSTR)&kwEnumPrivateData);
      IFSCloseFile(kwEnumPrivateData.hKWData);
    }
} // CollectInFile


/* CollectInDirectory
 *
 * List a Directory containing help files including all the subdirectories
 */
void CollectInDirectory(LPCSTR lpcszDirName, LPCSTR lpcszKeyWord)
{
      DIR           * dir=opendir(lpcszDirName);
      struct dirent * ent;
      HIFS            ifs;
      char            szFullPathName[MAX_PATH];
      LPCSTR          point;

      if (dir)
      {
        while (NULL!=(ent=readdir(dir)))
        {
            if (strcmp(lpcszDirName, "."))
                sprintf(szFullPathName, 
                    "%s/%s",
                    lpcszDirName,
                    ent->d_name);
            else
                strcpy(szFullPathName, ent->d_name);

            if (NULL!=(point=strrchr(ent->d_name, '.')) &&
                (0==  lstrcmpi(point+1, "hlp")||0==lstrcmpi(point+1, "mvb")) &&
                NULL!=(ifs = IFSOpen(szFullPathName))
                )
            {
              CollectInFile(ifs, szFullPathName, lpcszKeyWord);

              IFSClose(ifs);
            }
            else
              if (ent->d_name[0]!='.')
                CollectInDirectory(szFullPathName, lpcszKeyWord);
        }

        closedir(dir);
      } // if (dir)

} // CollectInDirectory




/* ListCollectedData
 *
 * List the Data collected as a table
 */
void ListCollectedData(void)
{
  int        iKwIndex,
             iTopicIndex;
  LPKWREF    lpKwRef;
  LPTOPICREF lpTopicRef;

  for (iKwIndex=0; iKwIndex<iKwRefCount; iKwIndex++)
    {
      lpKwRef = lpKwRefArray + iKwIndex;

      puts("<tr><td>");
      htmlputs(lpKwRef->lpcszKeyWord);
      puts("</td><td>");

      if (lpKwRef->iHitCount>1)
        puts("<ul>");

      for (iTopicIndex=0; iTopicIndex<lpKwRef->iHitCount; ++iTopicIndex)
        {
          lpTopicRef=lpKwRef->lpTopicRefs+iTopicIndex;
          if (lpKwRef->iHitCount>1)
            puts("<li>");
          printf("<a href=\"" CGI_NAME "?FILENAME=%s&TOPICOFFSET=0x%08x\">",
                 lpTopicRef->lpcszHelpFileName,
                 lpTopicRef->dwTopicOffset);
          htmlputs(lpTopicRef->lpcszTopicTitle);
          puts("</a>\n");
        }

      if (lpKwRef->iHitCount>1)
        puts("</ul>");

      puts("</td></tr>");
    }



} // ListCollectedData



/* SearchKeyWordInAllHelpFiles
 *
 * Search for the specified keyword in all available help files.
 */
void SearchKeyWordInAllHelpFiles(LPCSTR lpcszKeyWord)
{
  printf("content-Type: text/html\n\n"
         "<html><head><title>%s</title></head>"
         "<body><h1>%s</h1>"
         "<table><tr><th>%s</th><th>%s</th></tr>",
         _("Search Results"),
         _("Search Results"),
         _("KeyWord"),
         _("Topic(s)")
         );

  CollectInDirectory(".", lpcszKeyWord);
#ifdef __netware__
  CollectInDirectory("sys:public", lpcszKeyWord);
  CollectInDirectory("sys:login", lpcszKeyWord);
#endif

  ListCollectedData();

  puts("</table>"
       "</body></html>");
} // SearchKeyWordInAllHelpFiles

